package com.hz.pm.api.projectlib.manage;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.StrUtil;
import com.baomidou.mybatisplus.core.conditions.Wrapper;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.hz.pm.api.common.model.constant.MhUnitConst;
import com.hz.pm.api.projectlib.entity.MhSystemReplaceInfo;
import com.hz.pm.api.projectlib.entity.QxProject;
import com.hz.pm.api.projectlib.entity.QxProjectSystemReplaceInfos;
import com.hz.pm.api.projectlib.entity.QxPurchase;
import com.hz.pm.api.projectlib.model.dto.QxProjectSystemReplaceInfosDTO;
import com.hz.pm.api.projectlib.model.enumeration.MhSystemReplaceType;
import com.hz.pm.api.projectlib.model.enumeration.QxProjectStatus;
import com.hz.pm.api.projectlib.model.req.*;
import com.hz.pm.api.projectlib.model.vo.QxProjectDetailVO;
import com.hz.pm.api.projectlib.model.vo.QxProjectSystemReplaceInfosVO;
import com.hz.pm.api.projectlib.model.vo.QxProjectVO;
import com.hz.pm.api.projectlib.model.vo.QxPurchaseVO;
import com.hz.pm.api.projectlib.service.IMhSystemReplaceInfoService;
import com.hz.pm.api.projectlib.service.IQxProjectService;
import com.hz.pm.api.projectlib.service.IQxProjectSystemReplaceInfosService;
import com.hz.pm.api.projectlib.service.IQxPurchaseService;
import com.hz.pm.api.user.helper.MhUnitCache;
import com.hz.pm.api.user.security.model.UserInfoDetails;
import com.hz.pm.api.user.util.LoginUserUtil;
import com.ningdatech.basic.exception.BizException;
import com.ningdatech.basic.model.PageVo;
import com.ningdatech.basic.util.CollUtils;
import lombok.RequiredArgsConstructor;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * <p>
 * QxProjectManage
 * </p>
 *
 * @author WendyYang
 * @since 10:42 2024/8/31
 */
@Component
@RequiredArgsConstructor
public class QxProjectManage {

    private final MhUnitCache mhUnitCache;
    private final IQxProjectService qxProjectService;
    private final IMhSystemReplaceInfoService mhSystemReplaceInfoService;
    private final IQxProjectSystemReplaceInfosService qxProjectSystemReplaceInfosService;
    private final IQxPurchaseService qxPurchaseService;

    private void checkExistProject(String projectName, Long buildOrgCode) {
        LambdaQueryWrapper<QxProject> query = Wrappers.lambdaQuery(QxProject.class)
                .eq(QxProject::getProjectName, projectName)
                .eq(QxProject::getBuildOrgCode, buildOrgCode);
        if (qxProjectService.count(query) > 0) {
            throw BizException.wrap("项目已存在");
        }
    }

    @Transactional(rollbackFor = Exception.class)
    public synchronized void submitProjectRecord(QxProjectRecordReq req) {
        checkExistProject(req.getProjectName(), req.getBuildOrgCode());
        QxProject project = new QxProject();
        project.setStatus(QxProjectStatus.DECLARED.getCode());
        BeanUtil.copyProperties(req, project);
        qxProjectService.save(project);
        if (CollUtil.isNotEmpty(req.getSourceSystemIds())) {
            List<String> sourceSystemIds = req.getSourceSystemIds();
            List<MhSystemReplaceInfo> systemReplaceInfos = mhSystemReplaceInfoService.listByIds(sourceSystemIds);
            Map<Boolean, List<MhSystemReplaceInfo>> replaceInfosMap = CollUtils.group(systemReplaceInfos,
                    w -> MhSystemReplaceType.IT.eq(w.getReplaceType()));
            List<QxProjectSystemReplaceInfos> projectSystemReplaceInfos = new ArrayList<>();
            List<MhSystemReplaceInfo> itReplaceInfos = replaceInfosMap.get(true);
            if (CollUtil.isNotEmpty(itReplaceInfos)) {
                Map<String, List<MhSystemReplaceInfo>> group = CollUtils.group(itReplaceInfos, MhSystemReplaceInfo::getTargetName);
                for (Map.Entry<String, List<MhSystemReplaceInfo>> entry : group.entrySet()) {
                    List<MhSystemReplaceInfo> value = entry.getValue();
                    QxProjectSystemReplaceInfos item = new QxProjectSystemReplaceInfos();
                    item.setProjectId(project.getId());
                    item.setReplaceType(value.get(0).getReplaceType());
                    item.setTargetSystemName(value.get(0).getTargetName());
                    item.setSourceSystemNames(CollUtils.joinByComma(value, MhSystemReplaceInfo::getSourceName));
                    item.setSourceSystemIds(CollUtils.joinByComma(value, MhSystemReplaceInfo::getId));
                    projectSystemReplaceInfos.add(item);
                }
            }
            List<MhSystemReplaceInfo> otherReplaceInfos = replaceInfosMap.get(false);
            if (CollUtil.isNotEmpty(otherReplaceInfos)) {
                for (MhSystemReplaceInfo mhSystemReplaceInfo : otherReplaceInfos) {
                    QxProjectSystemReplaceInfos item = new QxProjectSystemReplaceInfos();
                    item.setProjectId(project.getId());
                    item.setReplaceType(mhSystemReplaceInfo.getReplaceType());
                    item.setTargetSystemName(mhSystemReplaceInfo.getTargetName());
                    item.setSourceSystemNames(mhSystemReplaceInfo.getSourceName());
                    item.setSourceSystemIds(mhSystemReplaceInfo.getId());
                    projectSystemReplaceInfos.add(item);
                }
            }
            if (CollUtil.isNotEmpty(projectSystemReplaceInfos)) {
                qxProjectSystemReplaceInfosService.saveBatch(projectSystemReplaceInfos);
            }
        }
    }

    public synchronized void submitPurchaseRecord(QxPurchaseRecordReq req) {
        QxProject project = qxProjectService.getById(req.getProjectId());
        if (!QxProjectStatus.DECLARED.eq(project.getStatus())) {
            throw BizException.wrap("当前状态不允许采购备案");
        }
        QxPurchase purchase = qxPurchaseService.getByProjectId(req.getProjectId());
        if (purchase == null) {
            purchase = new QxPurchase();
        }
        BeanUtil.copyProperties(req, purchase);
        qxPurchaseService.saveOrUpdate(purchase);
    }

    public void submitAcceptRecord(QxAcceptRecordReq req) {
        QxProject project = qxProjectService.getById(req.getProjectId());
        if (!QxProjectStatus.UPGRADED.eq(project.getStatus())) {
            throw BizException.wrap("当前状态不允许验收备案");
        }
        Wrapper<QxPurchase> update = Wrappers.lambdaUpdate(QxPurchase.class)
                .eq(QxPurchase::getProjectId, req.getProjectId())
                .set(QxPurchase::getFinalAcceptDate, req.getFinalAcceptDate())
                .set(QxPurchase::getFinalAcceptFile, req.getFinalAcceptFile())
                .set(QxPurchase::getFirstAcceptDate, req.getFirstAcceptDate())
                .set(QxPurchase::getFirstAcceptFile, req.getFirstAcceptFile())
                .set(QxPurchase::getUpdateOn, LocalDateTime.now())
                .set(QxPurchase::getUpdateBy, LoginUserUtil.getUserId());
        qxPurchaseService.update(update);
    }

    public void submitUpgradeRecord(QxUpgradeRecordReq req) {
        QxProject project = qxProjectService.getById(req.getProjectId());
        if (!QxProjectStatus.PURCHASED.eq(project.getStatus())) {
            throw BizException.wrap("当前状态不允许改造提升备案");
        }
        List<QxProjectSystemReplaceInfosDTO> replaceInfos = req.getSystemReplaceInfos();
        List<QxProjectSystemReplaceInfos> projectSystemReplaceInfos = new ArrayList<>();
        for (QxProjectSystemReplaceInfosDTO replaceInfo : replaceInfos) {
            QxProjectSystemReplaceInfos qxReplaceInfo = new QxProjectSystemReplaceInfos();
            BeanUtil.copyProperties(replaceInfo, qxReplaceInfo);
            projectSystemReplaceInfos.add(qxReplaceInfo);
        }
        qxProjectSystemReplaceInfosService.updateBatchById(projectSystemReplaceInfos);
    }

    public QxProjectDetailVO getProjectDetail(Long projectId) {
        QxProject project = qxProjectService.getById(projectId);
        if (project == null) {
            return null;
        }
        QxProjectDetailVO projectDetail = new QxProjectDetailVO();
        QxProjectVO retProject = BeanUtil.copyProperties(project, QxProjectVO.class);
        projectDetail.setProject(retProject);
        QxPurchase purchase = qxPurchaseService.getByProjectId(projectId);
        if (purchase != null) {
            QxPurchaseVO retPurchase = BeanUtil.copyProperties(purchase, QxPurchaseVO.class);
            projectDetail.setPurchase(retPurchase);
        }
        List<QxProjectSystemReplaceInfos> systemReplaceInfos = qxProjectSystemReplaceInfosService.listByProjectId(projectId);
        if (CollUtil.isNotEmpty(systemReplaceInfos)) {
            List<QxProjectSystemReplaceInfosVO> retSystemReplaceInfos = BeanUtil.copyToList(systemReplaceInfos, QxProjectSystemReplaceInfosVO.class);
            projectDetail.setSystemReplaceInfos(retSystemReplaceInfos);
        }
        return projectDetail;
    }

    public PageVo<QxProjectVO> page(QxProjectLibReq req) {
        LambdaQueryWrapper<QxProject> query = Wrappers.lambdaQuery(QxProject.class)
                .eq(req.getProjectYear() != null, QxProject::getProjectYear, req.getProjectYear())
                .eq(req.getUnitStrip() != null, QxProject::getUnitStrip, req.getUnitStrip())
                .eq(req.getStatus() != null, QxProject::getStatus, req.getStatus())
                .like(StrUtil.isNotBlank(req.getProjectName()), QxProject::getProjectName, req.getProjectName())
                .like(StrUtil.isNotBlank(req.getBuildOrgName()), QxProject::getBuildOrgName, req.getBuildOrgName())
                .exists(StrUtil.isNotBlank(req.getReplaceType()),
                        "select 1 from qx_project_system_replace_infos " +
                                "where qx_project_system_replace_infos.project_id = qx_project.id " +
                                "and replace_type = {0}", req.getReplaceType());
        UserInfoDetails userDetail = LoginUserUtil.loginUserDetail();
        if (req.getBuildOrgCode() != null) {
            if (!userDetail.hasSuperAdmin()) {
                List<Long> viewUnitIds = mhUnitCache.getViewChildIdsRecursion(userDetail.getMhUnitId());
                if (!viewUnitIds.contains(req.getBuildOrgCode())) {
                    return PageVo.empty();
                }
            }
            query.in(QxProject::getBuildOrgCode, mhUnitCache.getViewChildIdsRecursion(req.getBuildOrgCode()));
        } else {
            if (!LoginUserUtil.loginUserDetail().hasSuperAdmin()
                    && MhUnitConst.ROOT_UNIT_ID != userDetail.getMhUnitId()) {
                query.in(QxProject::getBuildOrgCode, mhUnitCache.getViewChildIdsRecursion(userDetail.getMhUnitId()));
            }
        }
        Page<QxProject> page = qxProjectService.page(req.page(), query);
        return PageVo.of(BeanUtil.copyToList(page.getRecords(), QxProjectVO.class), page.getTotal());
    }

}
