package com.hz.pm.api.warning;

import javax.annotation.Resource;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.StopWatch;
import cn.hutool.core.map.MapUtil;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.Wrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.hz.pm.api.common.helper.UserInfoHelper;
import com.hz.pm.api.common.model.constant.BizConst;
import com.hz.pm.api.common.statemachine.event.ProjectStateChangeEvent;
import com.hz.pm.api.common.statemachine.event.TenderStateChangeEvent;
import com.hz.pm.api.common.util.EnvironmentUtil;
import com.hz.pm.api.projectlib.entity.PurchaseStatusChange;
import com.hz.pm.api.projectlib.model.entity.Project;
import com.hz.pm.api.projectlib.model.entity.ProjectStatusChange;
import com.hz.pm.api.projectlib.model.enumeration.WarningFlowTypeEnum;
import com.hz.pm.api.projectlib.model.enumeration.WarningNoticeTypeEnum;
import com.hz.pm.api.projectlib.model.enumeration.status.ProjectStatus;
import com.hz.pm.api.projectlib.service.IProjectService;
import com.hz.pm.api.projectlib.service.IProjectStatusChangeService;
import com.hz.pm.api.projectlib.service.IPurchaseStatusChangeService;
import com.hz.pm.api.scheduler.contants.TaskConstant;
import com.hz.pm.api.sys.enumeration.ProjectEarlyWarningStatusEnum;
import com.hz.pm.api.sys.manage.EarlyWarningManage;
import com.hz.pm.api.sys.model.entity.ProjectEarlyWarning;
import com.hz.pm.api.sys.model.entity.WflowEarlyWarningRecords;
import com.hz.pm.api.sys.service.IEarlyWarningRecordsService;
import com.hz.pm.api.sys.service.IProjectEarlyWarningService;
import com.hz.pm.api.user.helper.MhUnitCache;
import com.hz.pm.api.user.model.entity.UserInfo;
import com.hz.pm.api.user.model.enumeration.RoleEnum;
import com.hz.pm.api.user.service.IUserInfoService;
import com.ningdatech.basic.util.CollUtils;
import com.ningdatech.basic.util.StrPool;
import com.wflow.bean.entity.WflowEarlyWarning;
import com.wflow.enums.WarningRuleTypeEnum;
import com.wflow.service.IEarlyWarningService;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.flowable.engine.TaskService;
import org.flowable.task.api.Task;
import org.junit.Test;

import com.alibaba.fastjson.JSON;
import com.hz.pm.api.AppTests;
import com.hz.pm.api.projectdeclared.model.entity.ProjectDraft;
import com.hz.pm.api.projectdeclared.service.IProjectDraftService;
import org.springframework.beans.factory.annotation.Autowired;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @Classname DraftTest
 * @Description
 * @Date 2023/2/7 10:56
 * @Author PoffyZhang
 */
@Slf4j
public class DelayWarningTest extends AppTests {

    @Autowired
    private IEarlyWarningService earlyWarningService;
    @Autowired
    private IProjectService projectService;
    @Autowired
    private EarlyWarningManage earlyWarningManage;
    @Autowired
    private IProjectEarlyWarningService projectEarlyWarningService;
    @Autowired
    private IProjectStatusChangeService projectStatusChangeService;
    @Autowired
    private UserInfoHelper userInfoHelper;
    @Autowired
    private EnvironmentUtil environmentUtil;
    @Autowired
    private IEarlyWarningRecordsService earlyWarningRecordsService;
    @Autowired
    private TaskService taskService;
    @Autowired
    private IPurchaseStatusChangeService purchaseStatusChangeService;

    private ChronoUnit getTimeUnit() {
        if (environmentUtil.isDevEnv()) {
            return ChronoUnit.MINUTES;
        } else {
            return ChronoUnit.DAYS;
        }
    }

    @Test
    public void one(){

        log.info("=========== 滞后预警任务开始 ===========");
        StopWatch stopWatch = new StopWatch();
        stopWatch.start();
        // 1.查询 填报的 滞后预警规则 填报类型的 每个单位的 每个规则
        Wrapper<WflowEarlyWarning> query = Wrappers.lambdaQuery(WflowEarlyWarning.class)
                .eq(WflowEarlyWarning::getRuleType, WarningRuleTypeEnum.DELAY_WARNING.getCode())
                .eq(WflowEarlyWarning::getIsOpen, true);
        List<WflowEarlyWarning> warnings = earlyWarningService.list(query);
        for (WflowEarlyWarning warning : warnings) {
            long mhUnitId = warning.getMhUnitId();
            String rule = warning.getRule();
            if (StringUtils.isNotBlank(rule)) {
                JSONArray ruleArray = JSON.parseArray(rule);
                if (CollUtil.isNotEmpty(ruleArray)) {
                    ruleArray.forEach(r -> {
                        JSONObject rJson = JSON.parseObject(JSON.toJSONString(r));
                        Integer biz = rJson.getInteger("biz");
                        Integer delayTime = rJson.getInteger("delayTime");
                        String notice = rJson.getString("notice");
                        WarningFlowTypeEnum flowType = WarningFlowTypeEnum.getByCode(biz);
                        if (Objects.isNull(delayTime) || Objects.isNull(biz) || flowType == null) {
                            log.warn("规则数据 错误 :{}", rJson);
                            return;
                        }
                        String path = flowType.getPath();
                        // 查询 所有这个单位 所配置滞后规则节点 未提交的项目
                        Wrapper<Project> projectQuery = Wrappers.lambdaQuery(Project.class)
                                .eq(Project::getBuildOrgCode, mhUnitId)
                                .eq(Project::getNewest, Boolean.TRUE)
                                .eq(Project::getStatus, flowType.getProjectStatus().getCode());
                        List<Project> needNextProjects = projectService.list(projectQuery);
                        List<String> projectCodes = CollUtils.fieldList(needNextProjects, Project::getProjectCode);
                        Map<String, ProjectStatusChange> projectStatusChangeMap = MapUtil.newHashMap();
                        String stepName = null;
                        switch (flowType) {
                            // 已立项未采购 获取 立项时间
                            case APPROVED_PROJECT_NOT_PURCHASED:{
                                Map<String, ProjectStatusChange> approvedProjectNotPurchasedMap = projectStatusChangeService.listLastEventMap(projectCodes,
                                        ProjectStateChangeEvent.DECLARED_RECORD_PASS);
                                if (CollUtil.isNotEmpty(approvedProjectNotPurchasedMap)){
                                    projectStatusChangeMap.putAll(approvedProjectNotPurchasedMap);
                                    stepName = TaskConstant.StepName.PURCHASE;
                                }
                            }
                            break;
                            // 已采购未改造完成 获取合同签订时间
                            case PURCHASED_UNMODIFIED:{
                                Map<String, ProjectStatusChange> purchasedUnmodifiedMap = projectStatusChangeService.listLastEventMap(projectCodes,
                                        ProjectStateChangeEvent.SUBMIT_PURCHASE_CONTRACT_RECORD);
                                if (CollUtil.isNotEmpty(purchasedUnmodifiedMap)){
                                    projectStatusChangeMap.putAll(purchasedUnmodifiedMap);
                                    stepName = TaskConstant.StepName.ADAPT_MODIFY;
                                }
                            }
                            break;
                            // 已改造完成未终验 获取单位确认时间
                            case MODIFIED_NOT_FINAL_INSPECTION:{
                                Map<String, ProjectStatusChange> modifiedNotFinalInspectionMap = projectStatusChangeService.listLastEventMap(projectCodes,
                                        ProjectStateChangeEvent.SUBMIT_PURCHASE_ORG_CONFIRM);
                                if (CollUtil.isNotEmpty(modifiedNotFinalInspectionMap)){
                                    projectStatusChangeMap.putAll(modifiedNotFinalInspectionMap);
                                    stepName = TaskConstant.StepName.FINAL;
                                }
                            }
                            break;
                            default:
                                log.error("Unexpected value: {}", flowType);
                        }
                        // 需要首页监测预警展示的项目
                        // 获取项目 滞后(历时)时间
                        HashMap<String, LocalDateTime> touchTimeMap = MapUtil.newHashMap();
                        String finalStepName = stepName;
                        List<Project> needToWaringProjects = needNextProjects.stream()
                                .filter(p -> {
                                    ProjectStatusChange change = projectStatusChangeMap.get(p.getProjectCode());
                                    LocalDateTime touchTime = change == null ? null : change.getCreateOn();
                                    if (touchTime != null) {
                                        touchTimeMap.put(p.getProjectCode() + StrPool.DASH + biz, touchTime);
                                        long duration = getDurationTime(touchTime,LocalDateTime.now(),getTimeUnit());
                                        if (duration > delayTime) {
                                            // 项目 滞后预警
                                            projectEarlyWarningService.earlyWarning(p, WarningRuleTypeEnum.DELAY_WARNING.getCode(), com.wflow.enums.WarningNoticeTypeEnum.DELAY.getCode(), finalStepName);
                                            return Boolean.TRUE;
                                        }
                                    }
                                    return Boolean.FALSE;
                                }).collect(Collectors.toList());

                        for (Project needToWaringProject : needToWaringProjects) {
                            // 目前没有通知
                            // 如果去预警通知人 1.超级管理员 2.单位管理员 如果都包含 都要发
                            List<String> batchEmployees = Lists.newArrayList();
                            if (StringUtils.isNotBlank(notice) && notice.contains(WarningNoticeTypeEnum.SUPER_ADMIN.getCode().toString())) {
                                // 获取超级管理员用户信息
                                List<UserInfo> superAdminUsers = userInfoHelper.getEmployeeCodesByRoleEnum(RoleEnum.SUPER_ADMIN);
                                List<String> superAdminEmployeeCodes = superAdminUsers.stream().map(UserInfo::getEmployeeCode).collect(Collectors.toList());
                                batchEmployees.addAll(superAdminEmployeeCodes);
                            }
                            if (StringUtils.isNotBlank(notice) && notice.contains(WarningNoticeTypeEnum.UNIT_ADMIN.getCode().toString())) {
                                // 获取单位管理员用户信息
                                List<UserInfo> unitAdminUsers = userInfoHelper.getEmployeeCodesByRoleEnum(RoleEnum.COMPANY_MANAGER);
                                // 过滤出当前项目所属单位的单位管理员
                                List<String> unitAdminEmployeeCodes = unitAdminUsers.stream().filter(user -> Objects.equals(user.getMhUnitId().toString(), needToWaringProject.getBuildOrgCode()))
                                        .map(UserInfo::getEmployeeCode).collect(Collectors.toList());
                                batchEmployees.addAll(unitAdminEmployeeCodes);
                            }
                            // 保存到预警记录中
                            earlyWarningManage.doEarlyWarningDelay(warning.getNoticeMethod(), biz, needToWaringProject,
                                    WarningRuleTypeEnum.DELAY_WARNING.getCode(), ProjectEarlyWarningStatusEnum.DELAY_TIME.getCode(),
                                    path, touchTimeMap);
                        }
                    });
                }
            }
        }

        stopWatch.stop();
        log.info("=========== 滞后预警任务结束 耗时{}s", stopWatch.getTotalTimeSeconds());
    }

    @Test
    public void two(){
        List<Project> projects = projectService.list(Wrappers.lambdaQuery(Project.class)
                .eq(Project::getNewest, Boolean.TRUE));

        if (CollUtil.isEmpty(projects)) {
            return;
        }

        Set<String> projectSet = Sets.newHashSet();
        Map<String, Project> projectMap = projects.stream().filter(p -> projectSet.add(p.getProjectCode()))
                .collect(Collectors.toMap(Project::getProjectCode, p -> p));
        List<ProjectEarlyWarning> warnings = projectEarlyWarningService.list();

        // 删除 已经不存在的项目
        List<ProjectEarlyWarning> toRemove = warnings.stream()
                .filter(w -> !projectMap.containsKey(w.getProjectCode()))
                .collect(Collectors.toList());
        if (CollUtil.isNotEmpty(toRemove)) {
            List<Long> removeIds = toRemove.stream().map(ProjectEarlyWarning::getId)
                    .collect(Collectors.toList());
            projectEarlyWarningService.removeBatchByIds(removeIds);
        }

        //  去查询 各个报警 是否已经正常
        List<ProjectEarlyWarning> trueWarnings = warnings.stream().filter(w -> Objects.nonNull(w.getNormal()) && !w.getNormal()).collect(Collectors.toList());
        if (CollUtil.isEmpty(trueWarnings)) {
            return;
        }
        checkNormal(trueWarnings);
    }

    private long getDurationTime(LocalDateTime touchTime, LocalDateTime now, ChronoUnit timeUnit) {
        return timeUnit.between(touchTime, now);
    }

    // 3.去查询 各个报警 是否已经正常
    private void checkNormal(List<ProjectEarlyWarning> trueWarnings) {
        // 3.1 流程预警 如果正常了 要改会正常
        List<ProjectEarlyWarning> processWarning = trueWarnings.stream().filter(w -> Objects.nonNull(w.getProcessWarning())
                        && w.getProcessWarning())
                .collect(Collectors.toList());
        checkProcessWarning(processWarning);

        // 3.2 填报预警 如果正常了 要改会正常
        List<ProjectEarlyWarning> declaredWarning = trueWarnings.stream().filter(w ->
                        Objects.nonNull(w.getDeclaredWarning())
                                && w.getDeclaredWarning())
                .collect(Collectors.toList());
        checkDeclaredWarning(declaredWarning);

        // 3.3 滞后预警 如果正常了 要改会正常
        List<ProjectEarlyWarning> delayWarning = trueWarnings.stream().filter(w ->
                        Objects.nonNull(w.getDelayWarning())
                                && w.getDelayWarning())
                .collect(Collectors.toList());
        checkDelayWarning(delayWarning);
    }

    private void checkDelayWarning(List<ProjectEarlyWarning> delayWarning) {
        if (CollUtil.isEmpty(delayWarning)) {
            return;
        }
        for (ProjectEarlyWarning delay : delayWarning) {
            String projectCode = delay.getProjectCode();
            Project project = projectService.getProjectByCode(projectCode);
            if (Objects.isNull(project)) {
                continue;
            }
            WflowEarlyWarningRecords record = earlyWarningRecordsService.getOne(Wrappers.lambdaQuery(WflowEarlyWarningRecords.class)
                    .eq(WflowEarlyWarningRecords::getProjectCode, projectCode)
                    .eq(WflowEarlyWarningRecords::getRuleType, WarningRuleTypeEnum.DELAY_WARNING.getCode())
                    .orderByDesc(WflowEarlyWarningRecords::getWarningTime)
                    .last(BizConst.LIMIT_1));
            if (Objects.isNull(record)) {
                continue;
            }
            Integer biz = record.getBiz();
            WarningFlowTypeEnum flowTypeEnum = WarningFlowTypeEnum.getByCode(biz);
            if (Objects.isNull(flowTypeEnum)) {
                continue;
            }
            ProjectStatus projectStatus = flowTypeEnum.getProjectStatus();
            // 已经不是这个状态了 说明已经被提交了
            if (!projectStatus.eq(project.getStatus())) {
                checkCommited(delay, flowTypeEnum, projectCode);
                // 其它几种 都没有 说明已经正常了
                if (!delay.getProcessWarning() &&
                        !delay.getDeclaredWarning()) {
                    delay.setNormal(Boolean.TRUE);
                    delay.setStatus(ProjectEarlyWarningStatusEnum.NORMAL.name());
                }
                projectEarlyWarningService.updateById(delay);
            }
        }
    }

    private void checkDeclaredWarning(List<ProjectEarlyWarning> declaredWarnings) {
        if (CollUtil.isEmpty(declaredWarnings)) {
            return;
        }
        for (ProjectEarlyWarning declaredWarning : declaredWarnings) {
            String projectCode = declaredWarning.getProjectCode();
            Project project = projectService.getProjectByCode(projectCode);
            if (Objects.isNull(project)) {
                continue;
            }
            WflowEarlyWarningRecords record = earlyWarningRecordsService.getOne(Wrappers.lambdaQuery(WflowEarlyWarningRecords.class)
                    .eq(WflowEarlyWarningRecords::getProjectCode, projectCode)
                    .eq(WflowEarlyWarningRecords::getRuleType, WarningRuleTypeEnum.DECLARED_WARNING.getCode())
                    .orderByDesc(WflowEarlyWarningRecords::getWarningTime)
                    .last(BizConst.LIMIT_1));
            if (Objects.isNull(record)) {
                continue;
            }
            Integer biz = record.getBiz();
            WarningFlowTypeEnum flowTypeEnum = WarningFlowTypeEnum.getByCode(biz);
            if (Objects.isNull(flowTypeEnum)) {
                continue;
            }
            ProjectStatus projectStatus = flowTypeEnum.getProjectStatus();
            // 已经不是这个状态了 根据选择的节点判断是否完成了提交
            if (!projectStatus.eq(project.getStatus())) {
                checkCommited(declaredWarning, flowTypeEnum, projectCode);
                // 其它几种 都没有 说明已经正常了
                if (!declaredWarning.getProcessWarning() &&
                        !declaredWarning.getDelayWarning()) {
                    declaredWarning.setNormal(Boolean.TRUE);
                    declaredWarning.setStatus(ProjectEarlyWarningStatusEnum.NORMAL.name());
                }
                projectEarlyWarningService.updateById(declaredWarning);
            }
        }
    }

    private void checkCommited(ProjectEarlyWarning warning, WarningFlowTypeEnum flowTypeEnum, String projectCode) {
        switch (flowTypeEnum){
            // 已立项未采购 当用户采购&合同备案完成 说明进度正常
            case APPROVED_PROJECT_NOT_PURCHASED: {
                List<ProjectStateChangeEvent> events = Lists.newArrayList();
                events.add(ProjectStateChangeEvent.SUBMIT_PURCHASE_CONTRACT_RECORD);
                ProjectStatusChange projectStatusChange = projectStatusChangeService.getLastOne(projectCode, events);
                // 说明进行了采购&合同信息备案
                if (Objects.nonNull(projectStatusChange)) {
                    warning.setDeclaredWarning(Boolean.FALSE);
                }
            }
            break;
            // 已采购未改造完成 当用户适配改造完成已完成单位确认 说明进度正常
            case MODIFIED_NOT_FINAL_INSPECTION:{
                List<TenderStateChangeEvent> events = Lists.newArrayList();
                events.add(TenderStateChangeEvent.SUBMIT_PURCHASE_ORG_CONFIRM);
                PurchaseStatusChange purchaseStatusChange = purchaseStatusChangeService.getLastOne(projectCode, events);
                // 说明进行了单位确认
                if (Objects.nonNull(purchaseStatusChange)) {
                    warning.setDeclaredWarning(Boolean.FALSE);
                }

            }
            break;
            // 已改造完成未终验 当用户提交终验申请 说明进度正常
            case PURCHASED_UNMODIFIED:{
                List<ProjectStateChangeEvent> events = Lists.newArrayList();
                events.add(ProjectStateChangeEvent.FINAL_ACCEPTANCE_APPLICATION);
                ProjectStatusChange projectStatusChange = projectStatusChangeService.getLastOne(projectCode, events);
                // 说明提交了终验申请
                if (Objects.nonNull(projectStatusChange)) {
                    warning.setDeclaredWarning(Boolean.FALSE);
                }
            }
            break;
        }
    }

    /**
     * 流程 查看 是否已经处理过
     *
     * @param processWarnings
     */
    private void checkProcessWarning(List<ProjectEarlyWarning> processWarnings) {
        if (CollUtil.isEmpty(processWarnings)) {
            return;
        }
        for (ProjectEarlyWarning processWarning : processWarnings) {
            String projectCode = processWarning.getProjectCode();
            WflowEarlyWarningRecords record = earlyWarningRecordsService.getOne(Wrappers.lambdaQuery(WflowEarlyWarningRecords.class)
                    .eq(WflowEarlyWarningRecords::getProjectCode, projectCode)
                    .orderByDesc(WflowEarlyWarningRecords::getWarningTime)
                    .last(BizConst.LIMIT_1));
            if (Objects.isNull(record)) {
                continue;
            }
            String nodeId = record.getNodeId();
            if (StringUtils.isBlank(nodeId)) {
                continue;
            }
            List<Task> tasks = taskService.createTaskQuery()
                    .taskDefinitionKey(nodeId)
                    .list();
            if (CollUtil.isEmpty(tasks)) {
                //说明 已经审批通过了
                processWarning.setProcessWarning(Boolean.FALSE);
                //其它三种 都没有 说明已经正常了
                if (!processWarning.getDeclaredWarning() &&
                        !processWarning.getDelayWarning()) {
                    processWarning.setNormal(Boolean.TRUE);
                    processWarning.setStatus(ProjectEarlyWarningStatusEnum.NORMAL.name());
                }
                projectEarlyWarningService.updateById(processWarning);
            }
        }

    }

}
